#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.message import Message
from parlai.core.teachers import FixedDialogTeacher
from parlai.tasks.md_gender.build import build
from parlai.tasks.image_chat.agents import ImageChatTeacher as OrigImageTeacher
import parlai.tasks.md_gender.utils as gend_utils


from copy import deepcopy
import json
import os
import random


class ImageChatTeacher(FixedDialogTeacher):
    """
    Image Chat gender.
    """

    @staticmethod
    def add_cmdline_args(argparser):
        argparser = gend_utils.add_common_args(argparser)
        return argparser

    def __init__(self, opt, shared=None):
        self.opt = opt
        self.is_train = 'train' in opt['datatype'] and 'evalmode' not in opt['datatype']
        self.is_valid = 'valid' in opt['datatype']
        self.add_unknown_classes = opt['add_unknown_classes'] and self.is_train
        if shared is None:
            # set map
            self.data = self._setup_data(opt)
            if (self.is_train and opt['balance']) or (
                self.is_valid and opt['balance_valid']
            ):
                to_exclude = gend_utils.PARTNER_CANDS + gend_utils.SELF_CANDS
                self.data = gend_utils.balance_data(
                    self.data, key=1, exclude_labels=to_exclude
                )
        else:
            self.data = shared['data']
        super().__init__(opt, shared)

        self.label_candidates = gend_utils.ALL_CANDS

        self.reset()

    def _load_gender_data(self, opt):
        """
        Get data file.
        """
        build(opt)
        dt = opt['datatype'].split(':')[0]
        folder = os.path.join(
            opt['datapath'], 'md_gender', 'data_to_release', 'image_chat'
        )
        fle = 'engaging_imagechat_gender_captions_hashed.{}.jsonl'.format(dt)
        return os.path.join(folder, fle)

    def _setup_data(self, opt):
        # Load map from image ID to gender
        counts = {gend_utils.FEM: 0, gend_utils.MASC: 0, gend_utils.NEUTRAL: 0}
        self.image_id_to_gender = {}
        gender_data_fle = self._load_gender_data(opt)
        with open(gender_data_fle, 'r') as f:
            for line in f:
                gender_dct = json.loads(line)
                image_id = gender_dct['id']
                male = gender_dct['male']
                female = gender_dct['female']
                if not male and not female:
                    gender = gend_utils.NEUTRAL
                elif male and not female:
                    gender = gend_utils.MASC
                elif female and not male:
                    gender = gend_utils.FEM
                else:
                    # detected both a man and a woman in the photo
                    gender = gend_utils.NEUTRAL
                self.image_id_to_gender[image_id] = gender

        # Now load the captions
        image_teacher_opt = deepcopy(opt)
        image_teacher_opt['include_image'] = False
        orig_data = OrigImageTeacher(image_teacher_opt).data
        data = []
        missing_cnt = 0
        extra_data = []
        for ex in orig_data:
            image_num = ex['image_hash']
            gender = self.image_id_to_gender.get(image_num)
            if gender is None:
                missing_cnt += 1
                continue
            for dialog in ex['dialog']:
                text = dialog[1]
                label = f'ABOUT:{gender}'
                data.append((text, label, image_num, 'about'))
                counts[gender] += 1

                if self.add_unknown_classes:
                    extra_data.append(
                        (text, f'SELF:{gend_utils.NEUTRAL}', image_num, 'self')
                    )
                    extra_data.append(
                        (text, f'PARTNER:{gend_utils.NEUTRAL}', image_num, 'partner')
                    )

        # possibly sample the added unknown class data
        if len(extra_data) > 0:
            sample_rate = self.opt['unknown_temp']
            if sample_rate < 1.0:
                to_samp = int(sample_rate * len(extra_data))
                sampled = random.sample(extra_data, to_samp)
                data += sampled
            else:
                data += extra_data

        demarc = '=' * 50
        print('\n\n' + demarc)
        print('Gender breakdown:')
        tot = sum(counts.values())
        for k, v in counts.items():
            print(f'{k}: {v} ({v / tot})')
        print(f'Total: {tot}')
        print(demarc + '\n\n')

        if self.is_train:
            random.shuffle(data)

        return data

    def get(self, episode_idx, entry_idx=0):
        text, label, image_num, class_type = self.data[episode_idx]
        if class_type == 'self' or class_type == 'partner':
            # not TRUE neutral, sample all labels
            labels = gend_utils.UNKNOWN_LABELS[class_type]
        else:
            labels = [label]
        return Message(
            {
                'text': text,
                'labels': labels,
                'label_candidates': self.label_candidates[class_type],
                'episode_done': True,
                'id': 'ImageChat Gender',
                'image_id': image_num,
            }
        )

    def num_examples(self):
        return len(self.data)

    def num_episodes(self):
        return len(self.data)

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.message import Message
from parlai.core.teachers import FixedDialogTeacher
from parlai.tasks.md_gender.build import build
import parlai.tasks.md_gender.utils as gend_utils

import json
import os
import random

NEW_DATAFILE = 'md_gender/data_to_release/new_data/data.jsonl'


class MdGenderTeacher(FixedDialogTeacher):
    """
    MDGender data collected on Mechanical Turk.
    """

    @staticmethod
    def add_cmdline_args(argparser):
        argparser = gend_utils.add_common_args(argparser)
        agent = argparser.add_argument_group('New data gender')
        agent.add_argument(
            '--labels-to-use',
            type=str,
            default='all',
            choices=['all', 'self', 'partner', 'about'],
            help='Which labels to use for this teacher',
        )
        return argparser

    def __init__(self, opt, shared=None):
        self.opt = opt
        self.labels_to_use = opt['labels_to_use']
        self.is_train = 'train' in opt['datatype'] and 'evalmode' not in opt['datatype']
        self.is_valid = 'valid' in opt['datatype']
        self.label_candidates = gend_utils.ALL_CANDS

        if shared is None:
            # set map
            self.data = self._setup_data(opt)
            if (self.is_train and opt['balance']) or (
                self.is_valid and opt['balance_valid']
            ):
                to_exclude = (
                    ['ABOUT:non-binary'],
                )  # not enough non-binary examples to balance
                self.data = gend_utils.balance_data(
                    self.data, key='labels', exclude_labels=to_exclude
                )
        else:
            self.data = shared['data']
        super().__init__(opt, shared)
        self.reset()

    def _get_convos(self, opt):
        """
        Get the test/train/valid split.
        """
        build(opt)
        with open(os.path.join(opt['datapath'], NEW_DATAFILE), 'r') as f:
            ex_jsons = f.read().splitlines()

        convos = [json.loads(ex) for ex in ex_jsons]

        return convos

    def _setup_data(self, opt):
        # Load map from image ID to gender
        # TODO: proper train/test/valid splits
        data = []
        convos = self._get_convos(opt)
        for convo in convos:
            class_type = convo['class_type']
            if self.labels_to_use in ['all', class_type]:
                data.append(convo)

        if self.is_train:
            random.shuffle(data)

        gend_utils.get_data_stats(data, key='labels')

        return data

    def get(self, episode_idx, entry_idx=0):
        ex = self.data[episode_idx]
        class_type = ex['class_type']
        ex['label_candidates'] = gend_utils.ALL_CANDS[class_type]
        if 'unknown' in ex['labels'][0]:
            ex['labels'] = gend_utils.UNKNOWN_LABELS[class_type]
        return Message(ex)

    def num_examples(self):
        return len(self.data)

    def num_episodes(self):
        return len(self.data)

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import random
import numpy as np
import math
import json
import logging
from mephisto.core.registry import register_mephisto_abstraction
from mephisto.server.blueprints.static_react_task.static_react_blueprint import (
    StaticReactBlueprint,
)
from mephisto.core.argparse_parser import str2bool
from typing import Any, Dict, TYPE_CHECKING
from argparse import _ArgumentGroup as ArgumentGroup

if TYPE_CHECKING:
    from mephisto.data_model.task import TaskRun


def get_task_path():
    return os.path.dirname(__file__)


@register_mephisto_abstraction()
class TurnAnnotationsStaticBlueprint(StaticReactBlueprint):
    """
    This Blueprint has a subtasks number option to combine multiple conversations into
    "sub-HITs".

    It also has options for the onboarding data answers and the annotation bucket
    definitions.
    """

    BLUEPRINT_TYPE = 'turn_annotations_static_blueprint'

    def __init__(self, task_run: "TaskRun", opts: Any):
        super().__init__(task_run, opts)
        print(f'Running {self.__class__.__name__} with opts: {self.opts}')
        random.seed(self.opts["random_seed"])
        np.random.seed(self.opts["random_seed"])
        self.subtasks_per_unit = self.opts['subtasks_per_unit']
        self.conversation_count = self.opts['conversation_count']

        if self.subtasks_per_unit <= 0:
            raise Exception(
                f'subtasks-per-unit must be greater than zero but was {self.subtasks_per_unit}'
            )

        self.raw_data = self._initialization_data_dicts

        # Load from file if needed specifying which utterances within each
        # conversation to annotate
        self.annotation_indices = None
        if self.opts['annotation_indices_jsonl']:
            self.annotation_indices = []
            with open(
                self.opts['annotation_indices_jsonl'], "r", encoding="utf-8-sig"
            ) as f:
                line = f.readline()
                while line:
                    conversation_indices = json.loads(line)
                    self.annotation_indices.append(conversation_indices)
                    line = f.readline()
            if len(self.annotation_indices) != len(self.raw_data):
                raise Exception(
                    f'Cannot specify a different length of annotation indices ({len(self.annotation_indices)}) than conversations ({len(self.raw_data)}).'
                )
            # TODO: should check that utterances specified are all bot
            # utterances (agent_idx == 1)

        if self.conversation_count:
            self.raw_data = self.raw_data[: self.conversation_count]
            if self.annotation_indices:
                self.annotation_indices = self.annotation_indices[
                    : self.conversation_count
                ]

        # Reorganize the self-chat data
        self._initialization_data_dicts = self.process_data(
            self.raw_data, annotation_indices=self.annotation_indices
        )

        # Now chunk the data into groups of <num_subtasks>
        grouped_data = []
        logging.info(
            f'Raw data length: {len(self.raw_data)}. self.subtasks_per_unit: {self.subtasks_per_unit}'
        )
        for i in range(0, len(self._initialization_data_dicts), self.subtasks_per_unit):
            chunk = self._initialization_data_dicts[i : i + self.subtasks_per_unit]
            grouped_data.append(chunk)
        self._initialization_data_dicts = grouped_data
        # Last group may have less unless an exact multiple
        logging.info(
            f'Grouped data into {len(self._initialization_data_dicts)} tasks with {self.subtasks_per_unit} subtasks each.'
        )

    @classmethod
    def add_args_to_group(cls, group: "ArgumentGroup") -> None:
        """
        Adds required options for TurnAnnotationStaticBlueprint.
        """
        super().add_args_to_group(group)
        group.add_argument(
            "--random-seed",
            dest="random_seed",
            type=int,
            default=42,
            help="seed for random",
        )
        group.add_argument(
            "--annotation-question",
            dest="annotation_question",
            type=str,
            default='Does this comment require any annotations? (Check all that apply)',
            help="The string displayed above the checkboxes for each annotation in the task.",
        )
        group.add_argument(
            "--subtasks-per-unit",
            dest="subtasks_per_unit",
            type=int,
            default=-1,
            help="number of subtasks/comparisons to do per unit",
        )
        group.add_argument(
            "--annotation-indices-jsonl",
            dest="annotation_indices_jsonl",
            type=str,
            default=None,
            help="Specify which utterance indices to annotate per conversation in a JSONL file. Must be same length as conversations data-jsonl file. See example file in task_config/annotation_indices_example.jsonl",
        )
        group.add_argument(
            "--ask-reason",
            dest="ask_reason",
            type=str2bool,  # Need to handle it being 'False' in arg_string
            default=False,
            help="If we want to ask the crowdworker for a reason for each of their annotations in a text field",
        )
        group.add_argument(
            "--conversation-count",
            dest="conversation_count",
            type=int,
            default=None,
            help="Specify a positive integer if you want to use only the first N conversations in the data file",
        )
        group.add_argument(
            "--onboarding-data",
            dest="onboarding_data",
            type=str,
            default=os.path.join(get_task_path(), 'task_config/onboarding.json'),
            help="Path to data and answers for onboarding task in JSON format",
        )
        group.add_argument(
            "--annotation-buckets",
            dest="annotation_buckets",
            type=str,
            default=os.path.join(
                get_task_path(), 'task_config/annotation_buckets.json'
            ),
            help="As per Turn Annotations task, path to annotation buckets which will be checkboxes in the frontend for worker to annotate an utterance.",
        )

    def get_frontend_args(self) -> Dict[str, Any]:
        """
        Specifies what options within a task_config should be forwarded to the client
        for use by the task's frontend.
        """

        with open(self.opts['onboarding_data'], "r", encoding="utf-8-sig") as f:
            onboarding_data = json.loads(f.read())

        with open(self.opts['annotation_buckets'], "r", encoding="utf-8-sig") as f:
            annotation_buckets = json.loads(f.read())

        return {
            "task_description": self.opts['task_description'],
            "task_title": self.opts['task_title'],
            "annotation_question": self.opts['annotation_question'],
            "onboarding_data": onboarding_data,
            "annotation_buckets": annotation_buckets,
            "ask_reason": self.opts['ask_reason'],
            "frame_height": '100%',
            "num_subtasks": self.opts["subtasks_per_unit"],
            "block_mobile": True,
        }

    def process_data(self, data_dicts, annotation_indices=None):
        """
        Override this in a subclass if you want to change how data is processed from
        input file before being sent to the frontend.
        """
        output = []
        total_annotation_count = 0
        for conv_idx, d in enumerate(data_dicts):
            max_turn_to_show = len(d['dialog']) - 1
            if annotation_indices:
                total_annotation_count += len(annotation_indices[conv_idx])
                # We only want to show the conversation up to the last
                # utterance we need annotations on, b/c otherwise may confuse
                # or bias the turkers
                if len(annotation_indices[conv_idx]) > 1:
                    logging.info(
                        f'Splitting {len(annotation_indices[conv_idx])} separate problematic utterance annotations in the same conversation into two separate conversations for this task. This avoids biasing the turkers with utterances that may come after one of the annotations.'
                    )
                for a in annotation_indices[conv_idx]:
                    processed_dialog = self._process_conversation(d, [a])
                    output.append(processed_dialog)
            else:
                processed_dialog = self._process_conversation(d, [max_turn_to_show])
                output.append(processed_dialog)
        print(
            f'Processed {len(data_dicts)} total conversations into {len(output)} conversations to be used in crowdsourcing task with {total_annotation_count} total annotations.'
        )
        np.random.shuffle(output)
        return output

    def _process_conversation(self, d, annotation_indices):
        """
        Helper function for processing conversations.

        :param annotation_indices:
            Array of turn indices to annotate of the
            actual conversation not including the context [So 0 is the "Hi!" if
            that's the first non-context utterance of the conversation.]
        :return: modified dialogue object
        """
        new_dialogue = []
        max_turn_to_show = max(annotation_indices)
        adjusted_turn_idx = 0
        for full_turn in d['dialog']:
            if len(full_turn) != 2:
                print(
                    f'Warning! Skipping incomplete conversation! full_turn was: {full_turn}'
                )
                continue
            if adjusted_turn_idx > max_turn_to_show:
                logging.info(
                    f'Skipping {adjusted_turn_idx}th utterance, b/c max_turn_to_show was {max_turn_to_show}.'
                )
                continue
            # If there is a persona, which is context as in the ConvAI2
            # task, we don't want to display the persona utterances
            if 'persona' not in full_turn[0]['text']:
                do_annotate = False
                new_dialogue.append(
                    {
                        'text': full_turn[0]['text'],
                        'agent_idx': 0,
                        'do_annotate': do_annotate,
                        'other_metadata': full_turn[0]['other_metadata'],
                    }
                )
                adjusted_turn_idx += 1
            if 'persona' not in full_turn[1]['text']:
                do_annotate = True
                if annotation_indices:
                    do_annotate = adjusted_turn_idx in annotation_indices
                new_dialogue.append(
                    {
                        'text': full_turn[1]['text'],
                        'agent_idx': 1,
                        'do_annotate': do_annotate,
                        'other_metadata': full_turn[1]['other_metadata'],
                    }
                )
                adjusted_turn_idx += 1
        if adjusted_turn_idx < max_turn_to_show:
            raise Exception(
                f'Conversation had {adjusted_turn_idx} but max_turn_to_show was {max_turn_to_show}'
            )
        return new_dialogue


@register_mephisto_abstraction()
class TurnAnnotationsStaticInFlightQABlueprint(TurnAnnotationsStaticBlueprint):
    """
    This Blueprint mixes in a live onboarding as the last subtask (in addition to an
    onboarding at the start), and actually increases the number of subtasks per unit by
    1.
    """

    BLUEPRINT_TYPE = 'turn_annotations_static_inflight_qa_blueprint'

    def __init__(self, task_run: "TaskRun", opts: Any):
        super().__init__(task_run, opts)

        raw_qc_convos = []
        with open(self.opts['onboarding_in_flight_data'], "r") as f:
            line = f.readline()
            while line:
                qc_convo = json.loads(line)
                raw_qc_convos.append(qc_convo)
                line = f.readline()
        # Annotate all the utterances in the quality controls
        # So annotation_indices=None here
        self.quality_control_convos = self.process_data(
            raw_qc_convos, annotation_indices=None
        )

        # Re-chunk the data to add a quality control convo as the last subtask
        # No shuffling of the data for reproducibility's sake
        # (quality control will always be last subtask)
        # TODO: I don't think we need to re-chunk this actually; just iterate
        # over the data and add the quality control task
        all_data = []
        for grp in self._initialization_data_dicts:
            all_data.extend(grp)

        grouped_data = []
        number_of_tasks = math.floor(len(all_data) / self.subtasks_per_unit)
        for i in range(0, number_of_tasks):
            data_index = i * self.subtasks_per_unit
            chunk = all_data[data_index : data_index + self.subtasks_per_unit]
            qc_convo_idx = i % len(self.quality_control_convos)
            chunk.append(self.quality_control_convos[qc_convo_idx])
            grouped_data.append(chunk)

        self._initialization_data_dicts = grouped_data
        self.subtasks_per_unit = len(chunk)

        print(
            f'{self.__class__.__name__}: Grouped data into {len(self._initialization_data_dicts)} tasks with {self.subtasks_per_unit} subtasks each (added in-flight qualification task).'
        )

    @classmethod
    def add_args_to_group(cls, group: "ArgumentGroup") -> None:
        """
        Adds required options for TurnAnnotationsStaticInFlightQABlueprint.
        """
        super().add_args_to_group(group)
        group.add_argument(
            "--onboarding-in-flight-data",
            dest="onboarding_in_flight_data",
            type=str,
            default=os.path.join(
                get_task_path(), 'task_config/onboarding_in_flight.jsonl'
            ),
            help="Path to data and answers for onboarding task in JSON-L format (one JSON object per line per onboarding)",
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.message import Message
from parlai.core.teachers import FixedDialogTeacher
from parlai.tasks.convai2.agents import BothTeacher as OrigConvai2Teacher
from parlai.tasks.md_gender.build import build
from parlai.utils.misc import warn_once

import parlai.tasks.md_gender.utils as gend_utils

from copy import deepcopy
import json
import random
import os


gender_convert = {'man': 'male', 'woman': 'female', 'neutral': 'unknown', None: None}


def convert_text(text):
    """
    Normalize text.
    """
    new_text = text.lower()
    for x in gend_utils.CONTRACTION_SPACES:
        if x[1] in text:
            new_text = new_text.replace(x[1], x[0])
    for x in gend_utils.CONTRACTION_LEFT_SPACES:
        if x[1] in text:
            new_text = new_text.replace(x[1], x[0])
    for x in gend_utils.CONTRACTION_RIGHT_SPACES:
        if x[1] in text:
            new_text = new_text.replace(x[1], x[0])

    return new_text


class Convai2Teacher(FixedDialogTeacher):
    """
    Predict the gender of character given the dialogue utterance.
    """

    @staticmethod
    def add_cmdline_args(argparser):
        argparser = gend_utils.add_common_args(argparser)
        agent = argparser.add_argument_group('ConvAI2 gender')
        agent.add_argument(
            '--convai2-use-probably',
            type='bool',
            default=True,
            help='Use the probable gender of the persona as '
            'determined by crowdworkers, instead of the '
            'definitive one',
        )
        agent.add_argument(
            '--labels-to-use',
            type=str,
            default='all',
            choices=['all', 'self', 'partner'],
            help='Which labels to use for this teacher',
        )

    def __init__(self, opt, shared=None):
        self.opt = opt
        self.fixed_random = random.Random(42)
        self.label_candidates = [gend_utils.MASC, gend_utils.FEM, gend_utils.NEUTRAL]
        self.use_probably = opt['convai2_use_probably']
        self.labels_to_use = opt['labels_to_use']
        self.is_train = 'train' in opt['datatype'] and 'evalmode' not in opt['datatype']
        self.is_valid = 'valid' in opt['datatype']
        self.add_unknown_classes = opt['add_unknown_classes'] and self.is_train

        if shared and 'data' in shared:
            self.data = shared['data']
            self.persona_map = shared['persona_map']
        else:
            self.missing_cnt = 0
            self._load_persona_map(opt)
            self._setup_data(opt)
            if (self.is_train and opt['balance']) or (
                self.is_valid and opt['balance_valid']
            ):
                exclude_lst = gend_utils.ABOUT_CANDS
                self.data = gend_utils.balance_data(
                    self.data, exclude_labels=exclude_lst
                )

        self.label_candidates = gend_utils.ALL_CANDS

        opt = deepcopy(opt)
        super().__init__(opt, shared)
        self.reset()

    def _load_persona_map(self, opt):
        build(opt)
        persona_map_path = os.path.join(
            opt['datapath'],
            'md_gender',
            'data_to_release',
            'convai2',
            'convai2_all_personas_map.json',
        )
        with open(persona_map_path, 'rb') as f:
            self.persona_map = json.load(f)

    def num_episodes(self):
        return len(self.data)

    def num_examples(self):
        return len(self.data)

    def get_gender_annotations(self, lst):
        """
        Take a list of personas and find the gender.
        """
        new_lst = sorted([convert_text(persona) for persona in lst])
        persona_key = ' '.join(new_lst)
        if persona_key in self.persona_map:
            return self.persona_map[persona_key]
        else:
            return {'gender': 'neutral', 'probably': None, 'missing': True}

    def get_genders(self, your_persona, partner_persona):
        partner_persona_gender = self.get_gender_annotations(partner_persona)
        your_persona_gender = self.get_gender_annotations(your_persona)
        if 'missing' in partner_persona_gender:
            self.missing_cnt += 1
        if 'missing' in your_persona_gender:
            self.missing_cnt += 1

        partner_prob = None
        your_prob = None
        if self.use_probably:
            partner = partner_persona_gender['probably']
            if partner is None:
                partner = partner_persona_gender['gender']
            your = your_persona_gender['probably']
            if your is None:
                your = your_persona_gender['gender']
        else:
            partner = partner_persona_gender['gender']
            your = your_persona_gender['gender']
            partner_prob = partner_persona_gender['probably']
            your_prob = your_persona_gender['probably']

        your = gender_convert[your]
        your_prob = gender_convert[your_prob]
        partner = gender_convert[partner]
        partner_prob = gender_convert[partner_prob]

        return your, your_prob, partner, partner_prob

    def _setup_data(self, opt):
        counts = {
            'partner': {gend_utils.UNKNOWN: 0, gend_utils.FEM: 0, gend_utils.MASC: 0},
            'self': {gend_utils.UNKNOWN: 0, gend_utils.FEM: 0, gend_utils.MASC: 0},
        }

        dt = opt['datatype'].split(':')[0]
        if dt == 'test':
            warn_once('No test set; switching to valid')
            dt = 'valid'

        # build data
        print('[ Building data ... ]')
        new_eps = []
        orig_teacher = OrigConvai2Teacher(opt)
        total_exs = orig_teacher.num_examples()
        num_exs = 0
        while num_exs < total_exs:
            current_episode = []
            episode_done = False

            while not episode_done:
                # TODO: eventually all teachers should return Messages, so
                # we should assert this
                action = Message(orig_teacher.act())
                current_episode.append(action)
                episode_done = action.get('episode_done', False)
                num_exs += 1

            # now we have the entire episode,... do something
            first_ex = current_episode[0]
            first_ex_text = []
            partner_persona = []
            your_persona = []
            for line in first_ex['text'].split('\n'):
                # NOTE: we flip "your" and "partner" here since we are taking the 'text'
                # field instead of the 'label'
                if 'partner\'s persona: ' in line:
                    your_persona.append(line.split('partner\'s persona: ')[1])
                elif 'your persona: ' in line:
                    partner_persona.append(line.split('your persona: ')[1])
                else:
                    first_ex_text.append(line)

            your, your_prob, partner, partner_prob = self.get_genders(
                your_persona, partner_persona
            )

            for i, ex in enumerate(current_episode):
                counts['self'][your] += 1
                counts['partner'][partner] += 1
                if i == 0:
                    text = '\n'.join(first_ex_text)
                else:
                    text = ex['text']
                new_ex = {
                    'text': text,
                    'episode_done': True,
                    'your_persona': '\n'.join(your_persona),
                    'partner_persona': '\n'.join(partner_persona),
                    'id': 'ConvAI2 Gender',
                }
                if not self.use_probably:
                    new_ex['partner_prob'] = partner_prob
                    new_ex['your_prob'] = your_prob

                if your is not None and self.labels_to_use != 'partner':
                    # Get the your task
                    labels = [f'SELF:{your}']
                    your_ex = deepcopy(new_ex)
                    your_ex['labels'] = labels
                    your_ex['class_type'] = 'self'
                    new_eps.append(your_ex)

                if partner is not None and self.labels_to_use != 'self':
                    # Get the partner task
                    labels = [f'PARTNER:{partner}']
                    partner_ex = deepcopy(new_ex)
                    partner_ex['labels'] = labels
                    partner_ex['class_type'] = 'partner'
                    new_eps.append(partner_ex)

        if self.labels_to_use == 'all' and self.add_unknown_classes:
            # load about data
            all_about_data = gend_utils.get_inferred_about_data(
                self.opt['task'], self.opt
            )
            sample_rate = self.opt['unknown_temp']
            if sample_rate < 1.0:
                to_samp = int(sample_rate * len(all_about_data))
                sampled = random.sample(all_about_data, to_samp)
                new_eps += sampled
            else:
                new_eps += all_about_data

        if self.is_train:
            random.shuffle(new_eps)

        self.data = new_eps
        print(f'Missing cnt: {self.missing_cnt} / {len(self.data) * 2}')
        for x in ['self', 'partner']:
            print(f'Totals for {x}:')
            subtot = sum(counts[x].values())
            for k, v in counts[x].items():
                print(f'\t{k}: {v} ({v / subtot})')

    def get(self, episode_idx, entry_idx):
        ep = self.data[episode_idx]
        # add label candidates
        # we do not want to hardcode these in the saved data, in case
        # we try to change them
        class_type = ep['class_type']

        if gend_utils.UNKNOWN in ep['labels'][0]:
            # for unknown, we alternate M/F during training
            ep['labels'] = gend_utils.UNKNOWN_LABELS[class_type]

        ep['label_candidates'] = self.label_candidates[class_type]
        return Message(ep)

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        shared['persona_map'] = self.persona_map
        return shared



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Simple IR baselines.

We plan to implement the following variants:
Given an input message, either:
(i) find the most similar message in the (training) dataset and output the
  response from that exchange; or
(ii) find the most similar response to the input directly.
(iii) if label_candidates are provided, simply ranks them according to their
  similarity to the input message.

Currently only (iii) is used.

Additonally, TFIDF is either used (requires building a dictionary) or not,
depending on whether you train on the train set first, or not.
"""

import math
from collections.abc import Sequence
import heapq
import json

import parlai.utils.torch as torch_utils
from parlai.core.agents import Agent
from parlai.core.dict import DictionaryAgent
from parlai.utils.io import PathManager


class MaxPriorityQueue(Sequence):
    """
    Fixed-size priority queue keeping the max_size largest items.
    """

    def __init__(self, max_size):
        """
        Initialize priority queue.

        :param max_size: maximum capacity of priority queue.
        """
        self.capacity = max_size
        self.lst = []

    def add(self, item, priority):
        """
        Add element to the queue, with a separate priority if desired.

        Element will not be added if the queue is at capacity and the element
        has lower priority than the lowest currently in the queue.

        :param item:
            item to add to queue.
        :param priority:
            priority to use for item.
        """
        if len(self.lst) < self.capacity:
            heapq.heappush(self.lst, (priority, item))
        elif priority > self.lst[0][0]:
            heapq.heapreplace(self.lst, (priority, item))

    def __getitem__(self, key):
        """
        Get item at specified index.

        :param key: integer index into priority queue, 0 <= index < max_size.

        :returns: item stored at the specified index.
        """
        return sorted(self.lst)[key][1]

    def __len__(self):
        """
        Return length of priority queue.
        """
        return len(self.lst)


stopwords = {
    'i',
    'a',
    'an',
    'are',
    'about',
    'as',
    'at',
    'be',
    'by',
    'for',
    'from',
    'how',
    'in',
    'is',
    'it',
    'of',
    'on',
    'or',
    'that',
    'the',
    'this',
    'to',
    'was',
    'what',
    'when',
    'where',
    '--',
    '?',
    '.',
    "''",
    "''",
    "``",
    ',',
    'do',
    'see',
    'want',
    'people',
    'and',
    "n't",
    "me",
    'too',
    'own',
    'their',
    '*',
    "'s",
    'not',
    'than',
    'other',
    'you',
    'your',
    'know',
    'just',
    'but',
    'does',
    'really',
    'have',
    'into',
    'more',
    'also',
    'has',
    'any',
    'why',
    'will',
}


def score_match(query_rep, text, length_penalty, dictionary=None):
    """
    Calculate the score match between the query representation the text.

    :param query_rep:
        base query representation to match text again.
    :param text:
        string to compare against query_rep for matching tokens
    :param length_penalty:
        scores are divided by the norm taken to this power
    :param dictionary:
        optional dictionary to use to tokenize text

    :returns:
        float score of match
    """
    if text == "":
        return 0
    words = [w for w in dictionary.tokenize(text.lower())]
    score = 0
    rw = query_rep['words']
    used = {}
    for w in words:
        if w in rw and w not in used:
            score += rw[w]
        used[w] = True
    norm = math.sqrt(len(used))
    norm = math.pow(norm * query_rep['norm'], length_penalty)
    if norm > 1:
        score /= norm
    return score


def rank_candidates(query_rep, cands, length_penalty, dictionary):
    """
    Rank candidates given representation of query.

    :param query_rep:
        base query representation to match text again.
    :param cands:
        strings to compare against query_rep for matching tokens
    :param length_penalty:
        scores are divided by the norm taken to this power
    :dictionary:
        dictionary to use to tokenize text

    :returns:
        ordered list of candidate strings in score-ranked order
    """
    if True:
        mpq = MaxPriorityQueue(100)
        for c in cands:
            score = score_match(query_rep, c, length_penalty, dictionary)
            mpq.add(c, score)
        return list(reversed(mpq))
    else:
        cands = list(cands)
        score = [0] * len(cands)
        for i, c in enumerate(cands):
            score[i] = -score_match(query_rep, c, length_penalty, dictionary)
        r = [i[0] for i in sorted(enumerate(score), key=lambda x: x[1])]
        res = []
        for i in range(min(100, len(score))):
            res.append(cands[r[i]])
        return res


class IrBaselineAgent(Agent):
    """
    Information Retrieval baseline.
    """

    @staticmethod
    def add_cmdline_args(parser):
        """
        Add command line args specific to this agent.
        """
        parser = parser.add_argument_group('IrBaseline Arguments')
        parser.add_argument(
            '-lp',
            '--length_penalty',
            type=float,
            default=0.5,
            help='length penalty for responses',
        )
        parser.add_argument(
            '-hsz',
            '--history_size',
            type=int,
            default=1,
            help='number of utterances from the dialogue history to take use '
            'as the query',
        )
        parser.add_argument(
            '--label_candidates_file',
            type=str,
            default=None,
            help='file of candidate responses to choose from',
        )
        DictionaryAgent.add_cmdline_args(parser)

    def __init__(self, opt, shared=None):
        """
        Initialize agent.
        """
        super().__init__(opt)
        self.id = 'IRBaselineAgent'
        self.length_penalty = float(opt['length_penalty'])
        self.dictionary = DictionaryAgent(opt)
        self.opt = opt
        self.history = []
        self.episodeDone = True
        if opt.get('label_candidates_file'):
            f = open(opt.get('label_candidates_file'))
            self.label_candidates = f.read().split('\n')

    def reset(self):
        """
        Reset agent properties.
        """
        self.observation = None
        self.history = []
        self.episodeDone = True

    def observe(self, obs):
        """
        Store and remember incoming observation message dict.
        """
        self.observation = obs
        self.dictionary.observe(obs)
        if self.episodeDone:
            self.history = []
        if 'text' in obs:
            self.history.append(obs.get('text', ''))
        self.episodeDone = obs.get('episode_done', False)
        return obs

    def act(self):
        """
        Generate a response to the previously seen observation(s).
        """
        if self.opt.get('datatype', '').startswith('train'):
            self.dictionary.act()

        obs = self.observation
        reply = {}
        reply['id'] = self.getID()

        # Rank candidates
        cands = None
        if obs.get('label_candidates', False) and len(obs['label_candidates']) > 0:
            cands = obs['label_candidates']
        if hasattr(self, 'label_candidates'):
            # override label candidates with candidate file if set
            cands = self.label_candidates
        if cands:
            hist_sz = self.opt.get('history_size', 1)
            left_idx = max(0, len(self.history) - hist_sz)
            text = ' '.join(self.history[left_idx : len(self.history)])
            rep = self.build_query_representation(text)
            reply['text_candidates'] = rank_candidates(
                rep, cands, self.length_penalty, self.dictionary
            )
            reply['text'] = reply['text_candidates'][0]
        return reply

    def save(self, path=None):
        """
        Save dictionary tokenizer if available.
        """
        path = self.opt.get('model_file', None) if path is None else path
        if path:
            self.dictionary.save(path + '.dict')
            data = {}
            data['opt'] = self.opt
            torch_utils.atomic_save(data, path)
            with PathManager.open(path + '.opt', 'w') as handle:
                json.dump(self.opt, handle)

    def build_query_representation(self, query):
        """
        Build representation of query, e.g. words or n-grams.

        :param query: string to represent.

        :returns: dictionary containing 'words' dictionary (token => frequency)
                  and 'norm' float (square root of the number of tokens)
        """
        rep = {}
        rep['words'] = {}
        words = [w for w in self.dictionary.tokenize(query.lower())]
        rw = rep['words']
        used = {}
        for w in words:
            assert len(self.dictionary.freq) > 0
            rw[w] = 1.0 / (1.0 + math.log(1.0 + self.dictionary.freq[w]))
            used[w] = True
        rep['norm'] = math.sqrt(len(words))
        return rep



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.message import Message
from parlai.core.teachers import ChunkTeacher
from parlai.tasks.wikipedia.build import build
from parlai.utils.misc import warn_once
import parlai.tasks.md_gender.utils as gend_utils

from copy import deepcopy
import json
import os
import random
import re
import spacy
from typing import List


NLP = spacy.load('en_core_web_sm')

DEBUG = False


def check_if_person(text):
    """
    Using spacy, check if the title of the Wikipedia passage is a person.
    """
    doc = NLP(text)
    is_person = False
    ents = [ent for ent in doc.ents]
    for ent in ents:
        if ent.label_ == 'PERSON':
            is_person = True
    return is_person


def get_gender(text):
    """
    Determine gender by the count of referring pronouns in the biography (he, she,
    they).
    """
    he_count = len(re.findall(' he ', text.lower()))
    she_count = len(re.findall(' she ', text.lower()))
    they_count = len(re.findall(' they ', text.lower()))

    if he_count == max(he_count, she_count, they_count):
        return gend_utils.MASC
    elif she_count == max(he_count, she_count, they_count):
        return gend_utils.FEM
    else:
        nonbinary_count = len(re.findall(' non-binary ', text.lower()))
        if nonbinary_count > 0:
            return gend_utils.NONBINARY
        return gend_utils.NEUTRAL


class WikipediaTeacher(ChunkTeacher):
    """
    Wikipedia gender.
    """

    @staticmethod
    def add_cmdline_args(argparser):
        argparser = gend_utils.add_common_args(argparser)
        agent = argparser.add_argument_group('Wiki gender')
        agent.add_argument(
            '--class-task',
            type=str,
            default='all',
            choices=['single', 'all'],
            help='Rank against all possibilities vs. F/M/N/NB',
        )
        agent.add_argument(
            '--mask-gendered-words',
            type='bool',
            default=False,
            help='Mask explicitly gendered words.',
        )
        return argparser

    def __init__(self, opt, shared=None):
        if shared is None:
            # set map
            self.opt = opt
            self._set_chunk_idx_to_file(opt)
        else:
            self.chunk_idx_to_file = shared['chunk_idx_to_file']

        self.class_task = opt['class_task']
        self.is_train = 'train' in opt['datatype'] and 'evalmode' not in opt['datatype']
        self.is_valid = 'valid' in opt['datatype']
        self.add_unknown_classes = (
            opt['add_unknown_classes'] and self.is_train and self.class_task == 'all'
        )

        if self.class_task == 'all':
            self.label_candidates = gend_utils.ALL_CANDS
        elif self.class_task == 'single':
            self.label_candidates = {
                'about': [
                    gend_utils.FEM,
                    gend_utils.MASC,
                    gend_utils.NEUTRAL,
                    gend_utils.NONBINARY,
                ]
            }

        self.mask_gendered_words = opt['mask_gendered_words']
        if self.mask_gendered_words:
            male, female = gend_utils.get_explicitly_gendered_words(self.opt)
            self.gendered_list = male + female

        self.counts = {
            gend_utils.FEM: 0,
            gend_utils.MASC: 0,
            gend_utils.NEUTRAL: 0,
            gend_utils.NONBINARY: 0,
        }
        super().__init__(opt, shared)

    def _set_chunk_idx_to_file(self, opt):
        # download wikipedia data
        wiki_opt = deepcopy(opt)
        wiki_opt['task'] = 'wikipedia:full'
        build(wiki_opt)

        # now divide data into subfolders
        data_folder = self._get_data_folder()
        self.chunk_idx_to_file = {}
        i = 0
        for subdir in os.listdir(data_folder):
            if subdir != 'README.md':
                for fle in os.listdir(os.path.join(data_folder, subdir)):
                    self.chunk_idx_to_file[i] = os.path.join(data_folder, subdir, fle)
                    i += 1

    def _get_data_folder(self):
        return os.path.join(self.opt['datapath'], 'wikipedia/full/wiki_full_extracted')

    def get_num_samples(self, opt) -> int:
        """
        Return the number of samples given the datatype.
        """
        datatype = opt['datatype']
        if 'train' in datatype:
            return 12774693, 12774693
        elif 'valid' in datatype:
            return 7410, 7410
        else:
            return 7441, 7441

    def get_fold_chunks(self, opt) -> List[int]:  # type: ignore
        """
        Return a list of chunk IDs (integer).

        Given the datatype (train/test/valid), return the list of chunk IDs that
        correspond to that split.
        """
        datatype = opt['datatype']
        all_chunk_idxs = list(self.chunk_idx_to_file.keys())
        if DEBUG:
            print(f'Total chunks: {len(all_chunk_idxs)}')
        if 'train' in datatype:
            return all_chunk_idxs[:-10]
        elif 'valid' in datatype:
            return all_chunk_idxs[-10:-5]
        else:
            return all_chunk_idxs[-5:]

    def load_from_chunk(self, chunk_idx: int):
        """
        [Abstract] Given the chunk index, load examples from that chunk.

        Return a list of tuples. The function `_create_message` will take these tuples
        to form the Message object that is returned by the teacher.
        """
        output = []
        chunk_path = self.chunk_idx_to_file[chunk_idx]

        extra_data = []
        with open(chunk_path) as wf:
            for article_json in wf:
                article = json.loads(article_json)
                title = article['title']
                text = article['text']

                title = title.split(' (')[0]
                is_person = check_if_person(title)
                if not is_person:
                    continue

                gender = get_gender(text)

                label = f'ABOUT:{gender}'
                for par in text.split('\n'):
                    if par:
                        output.append((par, title, label, gender, 'about'))
                        self.counts[gender] += 1

                        if self.add_unknown_classes:
                            extra_data.append(
                                (
                                    par,
                                    title,
                                    f'SELF:{gend_utils.UNKNOWN}',
                                    gender,
                                    'self',
                                )
                            )
                            extra_data.append(
                                (
                                    par,
                                    title,
                                    f'PARTNER:{gend_utils.NEUTRAL}',
                                    gender,
                                    'partner',
                                )
                            )

        if len(extra_data) > 0:
            # possibly sample unknown classes
            sample_rate = self.opt['unknown_temp']
            if sample_rate < 1.0:
                to_samp = int(sample_rate * len(extra_data))
                sampled = random.sample(extra_data, to_samp)
                output += sampled
            else:
                output += extra_data

        if DEBUG:
            print('\n\nGender count update:')
            for k, v in self.counts.items():
                print(f'{k}: {v}')

        if (self.is_train and self.opt['balance']) or (
            self.is_valid and self.opt['balance_valid']
        ):
            exclude_lst = [
                f'ABOUT:{gend_utils.NONBINARY}',
                f'SELF:{gend_utils.UNKNOWN}',
                f'PARTNER:{gend_utils.NEUTRAL}',
            ]  # not enough of each of these examples to balance
            output = gend_utils.balance_data(output, key=2, exclude_labels=exclude_lst)

        if len(output) == 0:
            warn_once(f'CHUNK {chunk_idx} is empty')

        return output

    def _mask_gendered_words(self, text):
        return gend_utils.mask_gendered_words(text, self.gendered_list)

    def create_message(self, queue_output, entry_idx=0) -> 'Message':
        """
        [Abstract] Given the tuple output of the queue, return an act.
        """
        par, title, lbl, gender, class_type = queue_output
        if self.class_task == 'all':
            if class_type == 'self':
                labels = gend_utils.UNKNOWN_LABELS['self']  # Not True neutral
            else:
                labels = [lbl]
        elif self.class_task == 'single':
            labels = [gender]

        if self.mask_gendered_words:
            par = self._mask_gendered_words(par)

        return Message(
            {
                'text': par,
                'name': title,
                'labels': labels,
                'label_candidates': self.label_candidates[class_type],
                'episode_done': True,
                'id': 'Wikipedia Gender',
            }
        )

    def share(self):
        shared = super().share()
        shared['chunk_idx_to_file'] = self.chunk_idx_to_file
        return shared


def write_gender_to_file(open_file, loaded_data):
    prev = None
    for _, title, _, gender, _ in loaded_data:
        if title != prev:
            open_file.write(f'{title}\t{gender}\n')
            prev = title



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os

from parlai.core.opt import Opt
from parlai.core.teachers import ParlAIDialogTeacher
from parlai.tasks.bot_adversarial_dialogue.build import (
    build_dialogue_datasets,
    build_human_safety_eval_dataset,
    get_adversarial_dialogue_folder,
    get_human_safety_eval_folder,
)
import parlai.utils.logging as logging
from parlai.utils.io import PathManager
from parlai.utils.misc import str_to_msg
from parlai.tasks.dialogue_safety.agents import OK_CLASS, NOT_OK_CLASS

# Constants
SAFETY_DICT = {'safe': OK_CLASS, 'unsafe': NOT_OK_CLASS}


def _adversarial_dialogue_datapath(opt: Opt) -> str:
    """
    Return the filepath for the specified datatype of the specified adversarial dialogue
    task.
    """
    build_dialogue_datasets(opt)
    # Build the data if it doesn't exist.
    dt = opt['datatype'].split(':')[0]
    data_path = os.path.join(
        get_adversarial_dialogue_folder(opt['datapath']),
        'bot_adversarial_dialogue_datasets',
        dt + '.txt',
    )
    return data_path


class BotAdversarialDialogueTeacher(ParlAIDialogTeacher):
    """
    Teacher for bot_adversarial_dialogue.
    """

    def add_cmdline_args(argparser):
        agent = argparser.add_argument_group('Bot Adversarial Dialogue options')
        agent.add_argument('--bad-num-turns', type=int, default=-1)
        agent.add_argument(
            '--bad-speaker-to-eval',
            type=str,
            default='all',
            choices=['bot', 'human', 'all'],
        )
        agent.add_argument(
            '--bad-safety-mix',
            type=str,
            default='all',
            choices=['safe', 'unsafe', 'all'],
        )

    def __init__(self, opt, shared=None):
        opt['parlaidialogteacher_datafile'] = _adversarial_dialogue_datapath(opt=opt)
        super().__init__(opt, shared=shared)
        self.id = '{}.numTurns_{}_speakerToEval_{}_safetyMix_{}'.format(
            self.id,
            self.opt['bad_num_turns'],
            self.opt['bad_speaker_to_eval'],
            self.opt['bad_safety_mix'],
        )

    def _setup_data(self, path):
        logging.info(f"Loading ParlAI text data: {path}")

        self.episodes = []
        self.num_exs = 0
        eps = []
        with PathManager.open(path, newline='\n', encoding='utf-8') as read:
            for line_no, line in enumerate(read, 1):
                msg = str_to_msg(line.rstrip('\n'))
                if msg and 'eval_labels' in msg:
                    raise ValueError(
                        f"It looks like you've written eval_labels as a key in your "
                        f"data file. This is not appropriate; labels will be converted "
                        f"for you automatically. This is happening on Line {line_no} "
                        f"in {path}. The line is:\n\t{line}"
                    )
                if msg and 'text' not in msg:
                    raise ValueError(
                        f'ParlaiDialogTeacher requires a "text" field in every '
                        f'entry, but one is missing in Line {line_no} in {path}. '
                        f'The line is:\n\t{line}'
                    )
                if msg and 'labels' not in msg:
                    raise ValueError(
                        f'ParlaiDialogTeacher requires a "labels" field in every '
                        f'entry, but one is missing in Line {line_no} in {path}. '
                        f'The line is:\n\t{line}'
                    )

                if (
                    self.opt['bad_speaker_to_eval'] != 'all'
                    and self.opt['bad_speaker_to_eval'] != msg['speaker_to_eval']
                ):
                    continue
                if (
                    self.opt['bad_safety_mix'] != 'all'
                    and SAFETY_DICT[self.opt['bad_safety_mix']] != msg['labels'][0]
                ):
                    continue
                if self.opt['bad_num_turns'] > 0:
                    dialog = msg['text'].split('\n')
                    msg.force_set(
                        'text', '\n'.join(dialog[-self.opt['bad_num_turns'] :])
                    )
                if msg:
                    self.num_exs += 1
                    eps.append(msg)
                    if msg.get('episode_done', False):
                        self.episodes.append(eps)
                        eps = []
        if len(eps) > 0:
            # add last episode
            eps[-1].force_set('episode_done', True)
            self.episodes.append(eps)
        if len(self.episodes) == 1 and line_no > 100:
            logging.error(
                f'The data in {path} looks like one very long episode. If this '
                f'is intentional, you may ignore this, but you MAY have a bug in '
                f'your data.'
            )


def _human_safety_eval_datapath(opt: Opt) -> str:
    """
    Return the filepath for the specified datatype of the specified human evaluation
    task on bot adversarial dialogue.
    """
    build_human_safety_eval_dataset(opt)
    # Build the data if it doesn't exist.
    logging.info(
        f'The data for human safety evaluation is test set only '
        f'regardless of your chosen datatype, which is {opt["datatype"]} '
    )
    data_path = os.path.join(
        get_human_safety_eval_folder(opt['datapath']), 'human_safety_eval', 'test.txt'
    )
    return data_path


class HumanSafetyEvaluationTeacher(ParlAIDialogTeacher):
    """
    Teacher for human safety evaluation on bot adversarial dialogues.
    """

    def __init__(self, opt, shared=None):
        opt['parlaidialogteacher_datafile'] = _human_safety_eval_datapath(opt=opt)
        super().__init__(opt, shared=shared)


class DefaultTeacher(BotAdversarialDialogueTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import shlex
import shutil
import subprocess
import random
import logging

from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript
from mephisto.core.operator import Operator

from importlib import import_module

# Have to import this even though it's not directly used
from parlai.crowdsourcing.tasks.turn_annotations_static import (
    turn_annotations_blueprint,
)

from mephisto.utils.scripts import MephistoRunScriptParser
from mephisto.providers.mturk.utils.script_utils import direct_soft_block_mturk_workers

LAUNCH_FILE = None

# Blueprint import required though not used; this satisfies linting
_ = turn_annotations_blueprint

TASK_DIRECTORY = os.path.dirname(os.path.abspath(__file__))
FRONTEND_SOURCE_DIR = os.path.join(TASK_DIRECTORY, "webapp")
FRONTEND_BUILD_DIR = os.path.join(FRONTEND_SOURCE_DIR, "build")
STATIC_FILES_DIR = os.path.join(FRONTEND_SOURCE_DIR, "src", "static")


def setup_mephisto(launch_config):
    """
    Mephisto data is saved in:

    <datapath>/data/runs/NO_PROJECT/<project_id>/<task_run_id>/<assignment_id>/<agent_id>/data
    -- The NO_PROJECT and /data/ thing will fixed in later versions of Mephisto
    -- agent_id can be mapped to MTurk worker_id with workers table in sqlite3 db
    """
    parser = MephistoRunScriptParser()
    full_requester_name = launch_config.REQUESTER
    if 'sandbox' in launch_config.PROVIDER and launch_config.REQUESTER:
        full_requester_name = full_requester_name + '_sandbox'

    parser.set_defaults(
        **{
            'provider_type': launch_config.PROVIDER,
            'requester_name': full_requester_name,
            'datapath': launch_config.DATAPATH,
        }
    )

    architect_type, requester_name, db, args = parser.parse_launch_arguments()

    arg_string = (
        f"--blueprint-type {launch_config.BLUEPRINT_TYPE} "
        f"--architect-type {architect_type} "
        f"--requester-name {requester_name} "
        f'--task-title "\\"{launch_config.TASK_TITLE}\\"" '
        f'--task-description "\\"{launch_config.TASK_DESCRIPTION}\\"" '
        f'--task-name {launch_config.TASK_NAME} '
        f'--task-source "{TASK_DIRECTORY}/webapp/build/bundle.js" '
        f'--task-reward {launch_config.TASK_REWARD} '
        f'--subtasks-per-unit {launch_config.SUBTASKS_PER_UNIT} '
        f'--annotation-buckets {launch_config.ANNOTATION_BUCKETS} '
        f'--ask-reason {launch_config.ASK_REASON} '
        f'--task-tags chat,conversation,dialog,partner '
        # How many workers to do each assignment
        f"--units-per-assignment {launch_config.UNITS_PER_ASSIGNMENT} "
        # Maximum tasks a worker can do across all runs with task_name (0=infinite)
        f"--maximum-units-per-worker {launch_config.MAX_UNITS_PER_WORKER} "
        f"--allowed-concurrent 1 "
        f'--data-jsonl {launch_config.FILE_DATA_JSONL} '
        f'--extra-source-dir "{STATIC_FILES_DIR}" '
        f'--port 2222 '
        f'--onboarding-qualification turn-ann-s-onb '
        f"-use-onboarding True "
    )
    # Optional flags:
    try:
        arg_string += (
            f'--annotation-question "\\"{launch_config.ANNOTATION_QUESTION}\\"" '
        )
    except Exception:
        logging.info(f'Launch config {launch_config} had no ANNOTATION_QUESTION field')

    try:
        arg_string += (
            f'--annotation-indices-jsonl {launch_config.ANNOTATION_INDICES_JSONL} '
        )
    except Exception:
        logging.info(f'Launch config {launch_config} had no ANNOTATION_INDICES_JSONL')

    try:
        arg_string += f'--conversation-count {launch_config.CONVERSATION_COUNT} '
    except Exception:
        logging.info(f'Launch config {launch_config} had no CONVERSATION_COUNT')

    try:
        arg_string += f'--onboarding-data {launch_config.ONBOARDING_DATA} '
    except Exception:
        logging.info(f'Launch config {launch_config} had no ONBOARDING_DATA')

    print(arg_string)
    return db, arg_string


def build_task():
    """
    Build the task.
    """
    return_dir = os.getcwd()
    os.chdir(FRONTEND_SOURCE_DIR)
    if os.path.exists(FRONTEND_BUILD_DIR):
        shutil.rmtree(FRONTEND_BUILD_DIR)
    packages_installed = subprocess.call(["npm", "install"])
    if packages_installed != 0:
        raise Exception(
            "please make sure npm is installed, otherwise view "
            "the above error for more info."
        )

    webpack_complete = subprocess.call(["npm", "run", "dev"])
    if webpack_complete != 0:
        raise RuntimeError(
            "Webpack appears to have failed to build your "
            "frontend. See the above error for more information."
        )
    os.chdir(return_dir)


def block_workers(launch_config, opt, local_db, requester_name=None):
    """
    Use a block list to block undesired crowdsource workers (Mechanical Turkers for
    example).
    """
    if not requester_name:
        print(
            'Specified no requester_name, which is done when running locally. Skipping blocking...'
        )
        return

    block_list = launch_config.WORKER_BLOCK_LIST
    soft_block_qual_name = 'turn_annotations_static_no'
    print(
        f'About to soft block {len(block_list)} workers on {launch_config.PROVIDER} by giving qualification: {soft_block_qual_name}'
    )
    direct_soft_block_mturk_workers(
        local_db, block_list, soft_block_qual_name, requester_name
    )


def run_task(opt):
    """
    Note: launch_config opt should be something like:
    parlai.crowdsourcing.tasks.turn_annotations_static.launch_config.LaunchConfig
    """

    launch_config_file = opt.get('launch_config')
    launch_module = import_module(launch_config_file)
    launch_config = launch_module.LaunchConfig
    db, arg_string = setup_mephisto(launch_config)
    if 'sandbox' not in launch_config.PROVIDER:
        block_workers(opt, db, launch_config.REQUESTER)

    build_task()
    operator = Operator(db)
    operator.parse_and_launch_run_wrapper(shlex.split(arg_string), extra_args={})
    operator.wait_for_runs_then_shutdown()


def setup_args(parser=None):
    """
    This task takes a single argument which is a Python class specifying the launch
    config.
    """
    if parser is None:
        parser = ParlaiParser(True, False, 'Static Turn Annotations Task')
    parser.add_argument(
        '--launch-config',
        type=str,
        required=True,
        help='A Python config file with variables used in this script.',
    )
    return parser


class TurnAnnotationsStaticRunner(ParlaiScript):
    @classmethod
    def setup_args(cls):
        return setup_args()

    def run(self):
        return run_task(self.opt)


if __name__ == '__main__':
    random.seed(42)
    TurnAnnotationsStaticRunner.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.message import Message
from parlai.core.teachers import FixedDialogTeacher
from parlai.tasks.md_gender.build import build
import parlai.tasks.md_gender.utils as gend_utils

from copy import deepcopy
import json
import pickle
import os
import random

PERSONA_PATH = 'md_gender/data_to_release/light/personas.json'
LIGHT_DATA_PATH = 'md_gender/data_to_release/light/data/{}_convs.pkl'


class LIGHTTeacher(FixedDialogTeacher):
    """
    Predict the gender of character given the dialogue utterance.
    """

    @staticmethod
    def add_cmdline_args(argparser):
        argparser = gend_utils.add_common_args(argparser)
        agent = argparser.add_argument_group('LIGHT gender')
        agent.add_argument(
            '--labels-to-use',
            type=str,
            default='all',
            choices=['all', 'self', 'partner'],
            help='Which labels to use for this teacher',
        )

    def __init__(self, opt, shared=None):
        self.opt = opt
        self.fixed_random = random.Random(42)
        self.label_candidates = gend_utils.ALL_CANDS
        self.labels_to_use = opt['labels_to_use']
        self.is_train = 'train' in opt['datatype'] and 'evalmode' not in opt['datatype']
        self.is_valid = 'valid' in opt['datatype']
        self.add_unknown_classes = opt['add_unknown_classes'] and self.is_train

        if shared and 'data' in shared:
            self.data = shared['data']
        else:
            self._setup_data(opt)
            if (self.is_train and opt['balance']) or (
                self.is_valid and opt['balance_valid']
            ):
                to_exclude = gend_utils.ABOUT_CANDS
                self.data = gend_utils.balance_data(
                    self.data, exclude_labels=to_exclude
                )

        opt = deepcopy(opt)
        super().__init__(opt, shared)
        self.reset()

    def num_episodes(self):
        return len(self.data)

    def num_examples(self):
        return len(self.data)

    def _setup_data(self, opt):
        build(opt)
        datatype = opt['datatype']
        dt = datatype.split(':')[0]
        # Build a map from persona to gender
        persona_map = {}
        personas = json.load(open(os.path.join(opt['datapath'], PERSONA_PATH), 'rb'))[
            'old'
        ]
        for gender, lst in personas.items():
            for x in lst:
                persona_map[int(x['char_id'])] = {'name': x['name'], 'gender': gender}

        # Build a list of dialogue utterances and associated persona IDs
        light_world = pickle.load(
            open(os.path.join(opt['datapath'], LIGHT_DATA_PATH.format(dt)), 'rb')
        )
        utt_to_pers = []
        for x in light_world:
            for act in x['conv_info']['acts']:
                text = act['text']
                p_uid = act['id'].lower()
                self_char_id = None
                partner_char_id = None
                for y in x['conv_info']['characters']:
                    # if identifying self utterances, grab your own id
                    if y[0].lower() == p_uid:
                        self_char_id = y[1]['id']
                        self_name = y[0].lower()
                    # else grab the partner's id
                    elif y[0].lower() != p_uid:
                        partner_char_id = y[1]['id']
                        partner_name = y[0].lower()
                if self_char_id is not None and partner_char_id is not None:
                    utt_to_pers.append(
                        {
                            'text': text,
                            'self_id': self_char_id,
                            'self_name': self_name,
                            'partner_id': partner_char_id,
                            'partner_name': partner_name,
                        }
                    )

        self.data = []
        missing = 0

        counts = {
            'partner': {gend_utils.UNKNOWN: 0, gend_utils.FEM: 0, gend_utils.MASC: 0},
            'self': {gend_utils.UNKNOWN: 0, gend_utils.FEM: 0, gend_utils.MASC: 0},
        }

        for x in utt_to_pers:
            if x['self_id'] in persona_map and x['partner_id'] in persona_map:
                self_gender = persona_map[x['self_id']]['gender']
                partner_gender = persona_map[x['partner_id']]['gender']
                act = {
                    'text': x['text'].lower(),
                    'self_id': x['self_id'],
                    'partner_id': x['partner_id'],
                    'id': 'LIGHT Gender',
                    'episode_done': True,
                }
                if self_gender == gend_utils.NEUTRAL:
                    # not True neutral
                    self_gender = gend_utils.UNKNOWN
                if partner_gender == gend_utils.NEUTRAL:
                    # not True neutral
                    partner_gender = gend_utils.UNKNOWN
                if self_gender is not None and self.labels_to_use != 'partner':
                    labels = [f'SELF:{self_gender}']
                    self_act = deepcopy(act)
                    self_act['labels'] = labels
                    self_act['class_type'] = 'self'
                    self.data.append(self_act)
                if partner_gender is not None and self.labels_to_use != 'self':
                    labels = [f'PARTNER:{partner_gender}']
                    partner_act = deepcopy(act)
                    partner_act['labels'] = labels
                    partner_act['class_type'] = 'partner'
                    self.data.append(partner_act)

                counts['partner'][partner_gender] += 1
                counts['self'][self_gender] += 1
            else:
                missing += 1

        if self.labels_to_use == 'all' and self.add_unknown_classes:
            # load about data
            all_about_data = gend_utils.get_inferred_about_data(
                self.opt['task'], self.opt
            )
            sample_rate = self.opt['unknown_temp']
            if sample_rate < 1.0:
                # do something here
                to_samp = int(sample_rate * len(all_about_data))
                sampled = random.sample(all_about_data, to_samp)
                self.data += sampled
            else:
                self.data += all_about_data

        total = len(self.data)
        print(f'Total: {total}')
        for x in ['self', 'partner']:
            print(f'Totals for {x}:')
            subtot = sum(counts[x].values())
            for k, v in counts[x].items():
                print(f'\t{k}: {v} ({v / subtot})')

    def get(self, episode_idx, entry_idx):
        ep = self.data[episode_idx]
        class_type = ep['class_type']
        if gend_utils.UNKNOWN in ep['labels'][0]:
            # during training, we flip between labels
            ep['labels'] = gend_utils.UNKNOWN_LABELS[class_type]
        ep['label_candidates'] = self.label_candidates[class_type]
        return Message(ep)

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import time
from dataclasses import dataclass, field
from typing import Any, List

import hydra
from mephisto.core.hydra_config import RunScriptConfig, register_script_config
from mephisto.core.operator import Operator
from mephisto.utils.scripts import load_db_and_process_config
from omegaconf import DictConfig

from parlai.crowdsourcing.tasks.acute_eval.acute_eval_blueprint import BLUEPRINT_TYPE

"""
Example script for running ACUTE-Evals.
The only argument that *must* be set for this to be run is:
``pairings_filepath``:  Path to pairings file in the format specified in the README.md

The following args are useful to tweak to fit your specific needs;
- ``annotations_per_pair``: A useful arg if you'd like to evaluate a given conversation pair more than once.
- ``num_matchup_pairs``: Essentially, how many pairs of conversations you would like to evaluate
- ``subtasks_per_unit``: How many comparisons you'd like a turker to complete in one HIT

"""

TASK_DIRECTORY = os.path.dirname(os.path.abspath(__file__))

defaults = [
    {"mephisto/blueprint": BLUEPRINT_TYPE},
    {"mephisto/architect": "local"},
    {"mephisto/provider": "mock"},
    {"conf": "example"},
]


@dataclass
class TestScriptConfig(RunScriptConfig):
    defaults: List[Any] = field(default_factory=lambda: defaults)
    task_dir: str = TASK_DIRECTORY
    current_time: int = int(time.time())


register_script_config(name='scriptconfig', module=TestScriptConfig)


@hydra.main(config_name="scriptconfig")
def main(cfg: DictConfig) -> None:
    db, cfg = load_db_and_process_config(cfg)
    operator = Operator(db)
    operator.validate_and_run_config(run_config=cfg.mephisto, shared_state=None)
    operator.wait_for_runs_then_shutdown(skip_input=True, log_rate=30)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.message import Message
from parlai.core.teachers import FixedDialogTeacher
import parlai.tasks.md_gender.utils as gend_utils
from parlai.tasks.md_gender.build import build

from copy import deepcopy
import json
import os
import random


class FunpediaTeacher(FixedDialogTeacher):
    """
    Funpedia Gender gender.
    """

    @staticmethod
    def add_cmdline_args(argparser):
        argparser = gend_utils.add_common_args(argparser)
        return argparser

    def __init__(self, opt, shared=None):
        self.opt = opt
        self.is_train = 'train' in opt['datatype'] and 'evalmode' not in opt['datatype']
        self.is_valid = 'valid' in opt['datatype']
        self.add_unknown_classes = opt['add_unknown_classes'] and self.is_train
        self.label_candidates = gend_utils.ALL_CANDS

        if shared is None:
            # set map
            self.data = self._setup_data(opt)
            if (self.is_train and opt['balance']) or (
                self.is_valid and opt['balance_valid']
            ):
                to_exclude = gend_utils.PARTNER_CANDS + gend_utils.SELF_CANDS
                self.data = gend_utils.balance_data(
                    self.data, key='labels', exclude_labels=to_exclude
                )
        else:
            self.data = shared['data']
        super().__init__(opt, shared)
        self.reset()

    def _load_gender_data(self, opt):
        build(opt)
        dt = opt['datatype'].split(':')[0]
        fle = os.path.join(
            opt['datapath'], 'md_gender', 'data_to_release', 'funpedia', f'{dt}.jsonl'
        )

        data = []
        with open(fle, 'r') as f:
            lines = f.read().splitlines()
            for line in lines:
                ex = json.loads(line)
                gender = ex['gender']
                ex['labels'] = [f'ABOUT:{gender}']
                ex['class_type'] = 'about'
                data.append(ex)

        return data

    def _setup_data(self, opt):
        # Load map from image ID to gender
        data = self._load_gender_data(opt)

        # Possibly add extra examples
        extra_data = []
        if self.add_unknown_classes:
            for ex in data:
                # add self examples
                self_ex = deepcopy(ex)
                self_ex['class_type'] = 'self'
                # not True neutral, so we flip between
                self_ex['labels'] = gend_utils.UNKNOWN_LABELS['self']
                extra_data.append(self_ex)
                # add partner examples
                partner_ex = deepcopy(ex)
                partner_ex['labels'] = gend_utils.UNKNOWN_LABELS['partner']
                partner_ex['class_type'] = 'partner'
                extra_data.append(partner_ex)

            # now sample the data
            sample_rate = self.opt['unknown_temp']
            if sample_rate < 1.0:
                to_samp = int(sample_rate * len(extra_data))
                sampled = random.sample(extra_data, to_samp)
                data += sampled
            else:
                data += extra_data

        if self.is_train:
            random.shuffle(data)

        gend_utils.get_data_stats(data, key='labels')

        return data

    def get(self, episode_idx, entry_idx=0):
        ex = self.data[episode_idx]
        class_type = ex['class_type']
        return Message(
            {
                'text': ex['text'],
                'labels': ex['labels'],
                'episode_done': True,
                'label_candidates': self.label_candidates[class_type],
                'id': 'Funpedia Gender',
                'title': ex['title'],
                'persona': ex['persona'],
            }
        )

    def num_examples(self):
        return len(self.data)

    def num_episodes(self):
        return len(self.data)

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


class LaunchConfig:
    """
    Use this file to put all the arguments (an alternative to launching them from the
    command line as params)
    """

    BLUEPRINT_TYPE = 'turn_annotations_static_inflight_qa_blueprint'

    # Run the below to register the requester if needed (first time only)
    # mephisto register mturk_sandbox --name:XXXX
    # --access-key-id:XXXX --secret-access-key:XXXX
    # To run locally, set REQUEST = None and PROVIDER = 'mock'
    REQUESTER = 'FIXME'
    PROVIDER = 'mturk_sandbox'

    # This datapath is where the database object goes
    # If not Mephisto data path below then requester register seems to do nothing
    DATAPATH = 'FIXME'

    TASK_TITLE = 'Chat with a fellow conversationalist'
    TASK_DESCRIPTION = '''<br>
    <b><h4>Task Description</h4></b>
    <br>
    Dummy Task Description.

    Lorem ipsum.
    <br><br>
      '''

    FILE_DATA_JSONL = 'FIXME'

    TASK_REWARD = 0.3
    SUBTASKS_PER_UNIT = 6

    # How many workers to do each assignment
    UNITS_PER_ASSIGNMENT = 5

    # Maximum tasks a worker can do across all runs with task_name (0=infinite)
    MAX_UNITS_PER_WORKER = 5

    # A list of worker IDs to block from doing the task.
    WORKER_BLOCK_LIST = []

    # Blueprint specific params
    ASK_REASON = False
    ANNOTATION_INDICES_JSONL = None
